#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "lion.h"
#include "misc.h"
#include "lfnmatch.h"

#define DEBUG_FLAG 1024
#include "debug.h"
#include "conf.h"
#include "parser.h"
#include "request.h"
#include "visited.h"

#include <gdbm.h>



static char *visited_dbfile = NULL;
static int   visited_mintime = 30;
static float visited_percent = 75.0;

// gdbm's file handle
static GDBM_FILE dbf = NULL;


// Because the way that Syabas gets videos, (open, lseek, read 16k, close)
// we never see "one long transfer" of media. This means we need to remember
// the last few transfers, and slowly tally up the byte count.

struct visited_struct {
	char *fname;
	bytes_t bytes;
	time_t ctime;
	time_t mtime;
	int seen;
};

typedef struct visited_struct visited_t;

#define VISITED_MAX 1000
THREAD_SAFE static visited_t visited_list[VISITED_MAX];

// Circular position index
THREAD_SAFE static int visited_insert_pos = 0;
THREAD_SAFE static int visited_search_pos = 0;
// Strictly not required, but speeds-up searches when less than MAX
THREAD_SAFE static int visited_num = 0;


visited_t *visited_db_find(char *fname);
int visited_db_update(visited_t *vnode);




int visited_init(void)
{
	int i;

	for (i = 0; i < VISITED_MAX; i++) {
		memset(&visited_list[i], 0, sizeof(visited_t));
	}

	visited_insert_pos = 0;
	visited_search_pos = 0;
	visited_num = 0;

    if (!dbf && visited_dbfile) {
        debugf("[visited] opening DB file '%s'\n", visited_dbfile);
        dbf = gdbm_open(visited_dbfile,
                        0,
                        GDBM_WRCREAT,
                        0600,
                        NULL);

        if (!dbf)
            debugf("[visited] failed to open DB file '%s': errno %d, gdbm_errno %d: %s\n",
                   visited_dbfile,
                   errno,
                   gdbm_errno,
                   gdbm_strerror(gdbm_errno));

    }

	return 0;
}



void visited_free(void)
{

	int i;

	for (i = 0; i < VISITED_MAX; i++) {
		SAFE_FREE(visited_list[ i ].fname);
	}

	visited_insert_pos = 0;
	visited_search_pos = 0;
	visited_num = 0;

    if (dbf) {

        gdbm_close(dbf);
        dbf = NULL;

    }

}





//
//
// CMD      : VISITED
// Required : FILE, MINTIME, PERCENT
// Optional :
//
void visited_cmd(char **keys, char **values,
			  int items,void *optarg)
{
	char *file, *mintime, *percent;

	// Required
	file    = parser_findkey(keys, values, items, "FILE");
	mintime = parser_findkey(keys, values, items, "MINTIME");
	percent = parser_findkey(keys, values, items, "PERCENT");

	if (!file || !*file || !mintime || !*mintime || !percent || !*percent) {
		printf("      : [visited] missing required field: FILE, MINTIME, PERCENT\n");
		//do_exit = 1;
		return;
	}

	SAFE_COPY(visited_dbfile, file);
	visited_mintime = atoi(mintime);
	visited_percent = atof(percent);

}



void visited_freenode(visited_t *vnode)
{
	SAFE_FREE(vnode->fname);
	memset(vnode, 0, sizeof(*vnode));
}




//
// Search for existing node matching this name, starting from
// "search_pos". (Looping at the end). Update search_pos to found location.
//
visited_t *visited_find(char *fname)
{
	int i, place;

	for (i = 0, place = visited_search_pos;
		 i < visited_num;
		 i++, place++) {

		//if (place >= VISITED_MAX) place = 0;
		if (place >= visited_num) place = 0;

		if (!mystrccmp(fname, visited_list[ place ].fname)) {
			visited_search_pos = place;
			return &visited_list[ place ];
		}
	}

    // Attempt to locate the node in the disk storage:
    return visited_db_find(fname);

}



// Use the next "insert_pos" node, clear it, add in new name, and ctime.
// return it.
visited_t *visited_add(char *fname)
{
	visited_t *vnode;

	// Already added? We do no longer check this here, we check it before
    // calling visited_add(). Otherwise we end up in a loop from db_add
	//if ((vnode = visited_find(fname))) return vnode;

	// insert_pos points to where we insert, free it first.
	vnode = &visited_list[ visited_insert_pos ];
	visited_freenode(vnode);

	// loop the list if we stepped over.
	if (visited_insert_pos++ >= VISITED_MAX) visited_insert_pos = 0;

	vnode->fname = strdup(fname);
	vnode->ctime = lion_global_time;

	if (visited_num < VISITED_MAX)
		visited_num++;

	debugf("[visited] added '%s'\n", fname);

	return vnode;
}



void visited_fileclosed(request_t *node)
{
	time_t duration;
	bytes_t in, out;
	float percent = 0.0;
	visited_t *vnode;

	// Quick reasons as to why we should ignore this file
	if (!node->althandle) return;
	if (!node->bytes_size) return;

	debugf("[visited] closed file '%s'\n", node->path);

	lion_get_bytes(node->althandle, &in, &out);

	// No bytes sent
	if (!in) return;

	duration = lion_get_duration(node->althandle);

	// Watched it for minimum time?
   	if (duration <= visited_mintime) return;

	// Calculate percentage bytes sent
	// look up the node, or create, so we can keep stats.
	vnode = visited_find(node->path);

	if (!vnode)
		vnode = visited_add(node->path);

	if (!node) return;


	// Update node.
	vnode->bytes += in;

	// Let's only update it once a second.
	if (vnode->mtime == lion_global_time) return;

	vnode->mtime = lion_global_time;


	//percent = node->bytes_size * 100 / vnode->bytes;
	percent = vnode->bytes * 100 / node->bytes_size;
	duration = vnode->mtime - vnode->ctime;

	debugf("[visited] file played for %ld seconds\n", duration);
	debugf("[visited] and we have sent %"PRIu64" bytes out of %"PRIu64". Bytes percent %2.2f%%\n",
		   vnode->bytes, node->bytes_size, percent );

	// Not watched enough
	if (percent < visited_percent) return;

	if (vnode->seen) return;

	// Update the DB here.
	debugf("[visited] adding to db '%s'\n", vnode->fname);

	vnode->seen = 1;

	// Write us to disk.
	visited_db_update(vnode);
}


int visited_file(request_t *node, lfile_t *file)
{
	visited_t *vnode;

	vnode = visited_find(file->name);
	if (!vnode) return 0;

	debugf("[visited] testing '%s'. %s found.\n", file->name,
		   vnode ? "" : "not");

	return vnode->seen;
}



int visited_db_update(visited_t *vnode)
{
    datum lore, data;
    int ret;

    if (!dbf) return 1; // No DB file open.

    // The key is the filename path
    lore.dptr  = vnode->fname;
    lore.dsize = strlen(vnode->fname) + 1; // Include the NULL in key.

    // Data is really just the "existance" but, we could just store it all
    data.dptr  = (void *)vnode;
    data.dsize = sizeof(*vnode);

    ret = gdbm_store(dbf, lore, data, GDBM_REPLACE);

    if (ret) {
        debugf("[visited] error inserting '%s' into DB: %d : %s\n",
               vnode->fname,
               gdbm_errno,
               gdbm_strerror(gdbm_errno));
        return 2;
    }

    debugf("[visited] Added '%s' to DB.\n", vnode->fname);

    return 0;
}

visited_t *visited_db_find(char *fname)
{
    datum lore, data;
    visited_t *result;
    visited_t tmp;

    if (!dbf) return NULL;

    lore.dptr  = fname;
    lore.dsize = strlen(fname) + 1;

    data = gdbm_fetch(dbf, lore);

    // Did we get something?
    if (!data.dptr || (sizeof(tmp) != data.dsize))
        return NULL;

    // We need to turn this into a real node.
    result = visited_add(fname);

    // Copy over the information.
    memcpy(&tmp, data.dptr, sizeof(tmp));

    result->seen  = tmp.seen;
    result->mtime = tmp.mtime;
    result->ctime = tmp.ctime;
    result->bytes = tmp.bytes;

    return result;
}


